#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import datetime
import getpass
import logging
import os
import re

from typing import Optional
from cryptography.hazmat.primitives.serialization import pkcs12

from .. import api, crypto, utils, error
from ..params import KeeperParams

from .base import GroupCommand, Command, dump_report_data, report_output_parser, field_to_title
from .enterprise import EnterpriseCommand

from ..proto import automator_pb2 as automator_proto
from ..proto import ssocloud_pb2 as ssocloud

automator_list_parser = argparse.ArgumentParser(prog='automator-list', parents=[report_output_parser])

automator_view_parser = argparse.ArgumentParser(prog='automator-view')
automator_view_parser.add_argument('target', help='Automator ID or Name.')

automator_create_parser = argparse.ArgumentParser(prog='automator-create')
automator_create_parser.add_argument('--name', dest='name', required=True, action='store', help='Automator name.')
automator_create_parser.add_argument('--node', dest='node', required=True, help='Node Name or ID.')


automator_edit_parser = argparse.ArgumentParser(prog='automator-edit')
automator_edit_parser.add_argument('target', help='Automator ID or Name.')
automator_edit_parser.add_argument('--name', dest='name', action='store', help='Automator name.')
automator_edit_parser.add_argument('--url', dest='url', action='store', help='Automator Webhook URL.')
automator_edit_parser.add_argument('--skill', dest='skill', action='append', choices=['device', 'team', 'team_for_user'],
                                   help='Automator Skills.')
automator_edit_parser.add_argument('--set', dest='setting', metavar="KEY=VALUE", action='append',
                                   help='Automator Settings. Use value: '
                                        '${file:filename} to load text file content. '
                                        '${base64:filename} to load binary file content and encode it with base64url. '
                                        '${env:environment_variable_name} to load the content of environment variable.')

automator_delete_parser = argparse.ArgumentParser(prog='automator-delete')
automator_delete_parser.add_argument('target', help='Automator ID or Name.')
automator_delete_parser.add_argument('--force', '-f', dest='force', action='store_true',
                                     help='Delete with no confirmation')

automator_action_parser = argparse.ArgumentParser(prog='automator-action')
automator_action_parser.add_argument('target', help='Automator ID or Name.')

automator_certificate_parser = argparse.ArgumentParser(prog='automator-certificate')
automator_certificate_parser.add_argument('file', help='PFX Certificate file path.')


def register_commands(commands):
    commands['automator'] = AutomatorCommand()


def register_command_info(_, command_info):
    command_info['automator'] = 'Manage Automator endpoints for SSO Cloud'


class AutomatorCommand(GroupCommand):
    def __init__(self):
        super(AutomatorCommand, self).__init__()
        self.register_command('list', AutomatorListCommand(), 'Displays a list of available automators.')
        self.register_command('view', AutomatorViewCommand(), 'Prints automator details.')
        self.register_command('create', AutomatorCreateCommand(), 'Creates automator endpoint.')
        self.register_command('edit', AutomatorEditCommand(), 'Changes automator configuration.')
        self.register_command('delete', AutomatorDeleteCommand(), 'Deletes automator.')
        self.register_command('setup', AutomatorSetupCommand(), 'Sets up automator.')
        action_command = AutomatorActionCommand()
        self.register_command('init', action_command, 'Initializes automator.')
        self.register_command('enable', action_command, 'Enables automator.')
        self.register_command('disable', action_command, 'Disables automator.')
        self.register_command('reset', action_command, 'Resets automator configuration to the default values.')
        action_command = AutomatorLogCommand()
        self.register_command('log', action_command, 'Retrieves automator logs.')
        self.register_command('log-clear', action_command, 'Clears automator logs.')
        self.register_command('certificate', AutomatorCertificateCommand(), 'Display certificate information.')

        self.default_verb = 'list'


class AutomatorMixin(object):
    @staticmethod
    def skill_to_name(skill):  # type: (automator_proto.SkillType) -> str
        if skill == automator_proto.DEVICE_APPROVAL:
            return 'Device Approval'
        elif skill == automator_proto.TEAM_APPROVAL:
            return 'Team Approval'
        elif skill == automator_proto.TEAM_FOR_USER_APPROVAL:
            return 'Team For User Approval'
        else:
            return str(skill)

    @staticmethod
    def setting_to_str(setting):  # type: (automator_proto.AutomatorSettingValue) -> str
        return '{0}={1}'.format(setting.settingTag, setting.settingValue or '')

    @staticmethod
    def dump_automators(params, fmt=None, filename=None):
        # type: (KeeperParams, Optional[str], Optional[str]) -> None
        table = []
        headers = ['id', 'name', 'node_id', 'enabled', 'url', 'skills']
        if fmt and fmt != 'json':
            headers = [field_to_title(x) for x in headers]
        if params.automators:
            for info in params.automators:
                row = [info.automatorId, info.name, info.nodeId, info.enabled, info.url,
                       [AutomatorMixin.skill_to_name(x.skillType) for x in info.automatorSkills]]
                table.append(row)
        return dump_report_data(table, headers=headers, fmt=fmt, filename=filename)

    @staticmethod
    def dump_automator(endpoint, status=False):    # type: (automator_proto.AutomatorInfo, bool) -> None
        logging.info('{0:>32s}: {1}'.format('Automator ID', endpoint.automatorId))
        logging.info('{0:>32s}: {1}'.format('Name', endpoint.name))
        logging.info('{0:>32s}: {1}'.format('Node ID', endpoint.nodeId))
        logging.info('{0:>32s}: {1}'.format('URL', endpoint.url))
        if endpoint.version:
            logging.info('{0:>32s}: {1}'.format('Version', endpoint.version))
        logging.info('{0:>32s}: {1}'.format('Enabled', 'Yes' if endpoint.enabled else 'No'))
        if status:
            logging.info('{0:>32s}: {1}'.format('Initialized', 'Yes' if endpoint.status.initialized else 'No'))
            if endpoint.status.initialized:
                delta = datetime.datetime.now() - datetime.datetime.fromtimestamp (endpoint.status.initializedTimestamp//1000)
                days = delta.days
                seconds = delta.seconds
                hours = seconds // 3600
                uptime = ''
                if days > 0:
                    uptime += f'{days} days '
                    if hours > 0:
                        uptime += f'{hours} hours'
                else:
                    minutes = (seconds % 3600) // 60
                    if hours > 0:
                        uptime += f'{hours} hours'
                        if minutes > 0:
                            uptime += f' {minutes} minutes'
                    else:
                        uptime += f' {minutes} minutes'
                logging.info('{0:>32s}: {1}'.format('Uptime', uptime))

            if endpoint.status.sslCertificateExpiration > 0:
                dt = datetime.datetime.fromtimestamp(endpoint.status.sslCertificateExpiration//1000)
                logging.info('{0:>32}: {1}'.format('Certificate Expires', dt.strftime('%x')))

            if endpoint.status.numberOfDevicesApproved > 0:
                logging.info('{0:>32}: {1}'.format('Approved Devices', endpoint.status.numberOfDevicesApproved))
            if endpoint.status.numberOfDevicesDenied > 0:
                logging.info('{0:>32}: {1}'.format('Denied Devices', endpoint.status.numberOfDevicesDenied))

            if endpoint.status.numberOfTeamsApproved > 0:
                logging.info('{0:>32}: {1}'.format('Approved Teams', endpoint.status.numberOfTeamsApproved))
            if endpoint.status.numberOfTeamsDenied > 0:
                logging.info('{0:>32}: {1}'.format('Denied Teams', endpoint.status.numberOfTeamsDenied))

            if endpoint.status.numberOfTeamMembershipsApproved > 0:
                logging.info('{0:>32}: {1}'.format('Approved Team Memberships', endpoint.status.numberOfTeamMembershipsApproved))
            if endpoint.status.numberOfTeamMembershipsDenied > 0:
                logging.info('{0:>32}: {1}'.format('Denied Team Memberships', endpoint.status.numberOfTeamMembershipsDenied))

            if endpoint.status.numberOfErrors > 0:
                logging.info('{0:>32}: {1}'.format('Number of Errors', endpoint.status.numberOfErrors))

        i = 0
        for skill in endpoint.automatorSkills:
            logging.info('{0:>32}: {1}'.format(
                'Skills' if i == 0 else '',
                AutomatorMixin.skill_to_name(skill.skillType)))
            i += 1

        visible_settings = [x for x in endpoint.automatorSettingValues if x.userVisible and x.editable]
        if visible_settings:
            logging.info('\n{0:>32s}\n'.format('Automator Settings'))
            for setting in visible_settings:
                setting_value = AutomatorMixin.setting_to_str(setting)
                value_len = len(setting_value)
                if value_len > 90:
                    value_len -= 80
                    setting_value = f'{setting_value[:80]}... {value_len} more characters.'

                logging.info('{0:>32s}: {1}'.format(setting.settingName, setting_value))

        if endpoint.logEntries:
            logging.info('\n{0:>32s}\n'.format('Automator Log'))
            for log in endpoint.logEntries:
                logging.info('<{0}> {1} - {2}'.format(log.messageLevel, log.serverTime, log.message))

    @staticmethod
    def ensure_loaded(params, force=False):   # type: (KeeperParams, bool) -> None
        if params.automators is None or force:
            enterprise_id = 0
            for node in params.enterprise['nodes']:
                enterprise_id = node['node_id'] >> 32
                break

            rq = automator_proto.AdminGetAutomatorsForEnterpriseRequest()
            rq.enterpriseId = enterprise_id
            admin_rs = api.communicate_rest(
                params, rq, 'automator/automator_get_all_for_enterprise', rs_type=automator_proto.AdminResponse)
            if admin_rs.success:
                params.automators = []
                params.automators.extend(admin_rs.automatorInfo)
            else:
                logging.warning(admin_rs.message)

    @staticmethod
    def find_automator(params, target):   # type: (KeeperParams, any) -> automator_proto.AutomatorInfo
        if not target:
            raise Exception('Automator Name cannot be empty')
        AutomatorMixin.ensure_loaded(params, False)
        if params.automators:
            try:
                target_id = int(target)
                for info in params.automators:
                    if target_id == info.automatorId:
                        return info
            except ValueError:
                pass

        name = target.lower() if type(target) == str else str(target)
        automators = [x for x in params.automators if x.name.lower() == name]
        if len(automators) == 0:
            raise Exception('Automator with name \"{0}\" not found'.format(target))
        if len(automators) > 1:
            raise Exception('There are %d automators with name \"{0})\". Use automator ID.'.format(target))
        return automators[0]


class AutomatorListCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):
        return automator_list_parser

    def execute(self, params, **kwargs):
        self.ensure_loaded(params, True)
        return self.dump_automators(params, fmt=kwargs.get('format'), filename=kwargs.get('output'))


class AutomatorViewCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):
        return automator_view_parser

    def execute(self, params, **kwargs):
        automator = self.find_automator(params, kwargs.get('target'))

        rq = automator_proto.AdminGetAutomatorRequest()
        rq.automatorId = automator.automatorId
        rs = api.communicate_rest(params, rq, 'automator/automator_get', rs_type=automator_proto.AdminResponse)

        self.dump_automator(rs.automatorInfo[0], status=True)
        if not rs.success:
            logging.info('')
            logging.info(rs.message)


class AutomatorCreateCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_create_parser

    def execute(self, params, **kwargs):
        # type: (KeeperParams, **any) -> any
        name = kwargs.get('name')
        if not name:
            logging.warning('\"--name\" option is required for \"create\" command')
            return
        node = kwargs.get('node')
        nodes = list(self.resolve_nodes(params, node))
        if len(nodes) == 0:
            logging.warning('Node name \'%s\' is not found', node)
            return
        if len(nodes) > 1:
            logging.warning('Node name \'%s\' is not unique. Use Node ID.', node)
            return
        matched_node_id = nodes[0]['node_id']
        self.ensure_loaded(params, False)
        # if params.automators:    # type: list[automator_proto.AutomatorInfo]
        #     n = next((True for x in params.automators if x.nodeId == matched_node_id), None)
        #     if n:
        #         logging.warning('Automator for node \"%s\" already exists', node)
        #         return

        rq = automator_proto.AdminCreateAutomatorRequest()
        rq.nodeId = matched_node_id
        rq.name = name
        rs = api.communicate_rest(params, rq, 'automator/automator_create', rs_type=automator_proto.AdminResponse)
        if rs.success:
            self.dump_automator(rs.automatorInfo[0])
            params.automators = None
        else:
            logging.warning(rs.message)


class AutomatorEditCommand(EnterpriseCommand, AutomatorMixin):
    parameter_pattern = re.compile(r'^\${([^:]+?):([^}]+?)}$')

    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_edit_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        rq = automator_proto.AdminEditAutomatorRequest()
        rq.automatorId = automator.automatorId
        rq.enabled = automator.enabled
        name = kwargs['name']
        if name:
            rq.name = name
        url = kwargs['url']   # type: str
        if url:
            if not url.startswith('http'):
                url = 'https://' + url
            if url.endswith('/'):
                url = url.rstrip('/')
            rq.url = url
        skills = kwargs.get('skill')
        if skills:
            for skill in skills:
                if skill == 'device':
                    rq.skillTypes.append(automator_proto.DEVICE_APPROVAL)
                elif skill == 'team':
                    rq.skillTypes.append(automator_proto.TEAM_APPROVAL)
                elif skill == 'team_for_user':
                    rq.skillTypes.append(automator_proto.TEAM_FOR_USER_APPROVAL)
                else:
                    logging.warning('Unsupported skill: \"%s\"', skill)
                    return
        settings = kwargs.get('setting')
        if settings:
            available_settings = {}
            for sv in automator.automatorSettingValues:
                available_settings[sv.settingName.lower()] = sv

            for setting in settings:
                pos = setting.find('=')
                if pos > 1:
                    key = setting[:pos].strip()
                    existing = available_settings.get(key.lower())
                    if not existing:
                        logging.warning('Invalid property name: \"%s\"', key)
                        return

                    setting_value = automator_proto.AutomatorSettingValue()
                    setting_value.settingId = existing.settingId
                    value = setting[pos+1:].strip()
                    if existing.dataType == ssocloud.INTEGER or existing.dataType == ssocloud.LONG:
                        try:
                            _ = int(value)
                        except ValueError:
                            logging.warning('Numeric property \"%s\" has incorrect value: \"%s\"', key, value)
                            return
                    elif existing.dataType == ssocloud.BOOLEAN:
                        if value.lower() in ['true', '1', 'on']:
                            value = 'true'
                        elif value.lower() in ['false', '0', 'off']:
                            value = 'false'
                        else:
                            logging.warning(
                                'Boolean property \"%s\" has incorrect value: \"%s\". Use \"true\" or \"false\"',
                                key, value)
                            return

                    match = AutomatorEditCommand.parameter_pattern.match(value.strip())
                    if match:
                        parts = match.groups()
                        if len(parts) == 2:
                            if parts[0].lower() == 'file':
                                filename = parts[1].strip()
                                if os.path.isfile(filename):
                                    with open(filename, 'r', encoding='utf-8') as f:
                                        value = f.read()
                            if parts[0].lower() == 'base64':
                                filename = parts[1].strip()
                                if os.path.isfile(filename):
                                    with open(filename, 'rb') as f:
                                        data = f.read()
                                        value = utils.base64_url_encode(data)
                            elif parts[0].lower() == 'env':
                                if parts[1] in os.environ:
                                    value = os.environ[parts[1]]

                    setting_value.settingValue = value
                    rq.automatorSettingValues.append(setting_value)
                else:
                    logging.warning('Incorrect setting format \"%s\". Expected KEY=VALUE', setting)
                    return

        rs = api.communicate_rest(params, rq, 'automator/automator_edit', rs_type=automator_proto.AdminResponse)
        if rs.automatorInfo:
            self.dump_automator(rs.automatorInfo[0])
            params.automators = None
        if not rs.success:
            logging.warning(rs.message)


class AutomatorSetupCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_action_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        rq = automator_proto.AdminSetupAutomatorRequest()
        rq.automatorId = automator.automatorId
        rq.automatorState = automator_proto.NEEDS_CRYPTO_STEP_1
        rs = api.communicate_rest(params, rq, 'automator/automator_setup',
                                  rs_type=automator_proto.AdminSetupAutomatorResponse)
        if not rs.success:
            logging.warning('Automator \"%s\" setup step #1 error: %s', automator.name, rs.message)
            return

        rq = automator_proto.AdminSetupAutomatorRequest()
        rq.automatorId = automator.automatorId
        rq.automatorState = automator_proto.NEEDS_CRYPTO_STEP_2
        automator_public_key = crypto.load_ec_public_key(rs.automatorEccPublicKey)
        keys = params.enterprise['keys']
        if 'ecc_encrypted_private_key' in keys:
            encrypted_ec_private_key = utils.base64_url_decode(keys['ecc_encrypted_private_key'])
            ec_private_key = crypto.decrypt_aes_v2(encrypted_ec_private_key, params.enterprise['unencrypted_tree_key'])
            encrypted_ec_private_key = crypto.encrypt_ec(ec_private_key, automator_public_key)
            rq.encryptedEccEnterprisePrivateKey = encrypted_ec_private_key

        if not params.forbid_rsa and 'rsa_encrypted_private_key' in keys:
            encrypted_rsa_private_key = utils.base64_url_decode(keys['rsa_encrypted_private_key'])
            rsa_private_key = crypto.decrypt_aes_v2(encrypted_rsa_private_key, params.enterprise['unencrypted_tree_key'])
            encrypted_rsa_private_key = crypto.encrypt_ec(rsa_private_key, automator_public_key)
            rq.encryptedRsaEnterprisePrivateKey = encrypted_rsa_private_key

        rq.encryptedTreeKey = crypto.encrypt_ec(params.enterprise['unencrypted_tree_key'], automator_public_key)
        rs = api.communicate_rest(params, rq, 'automator/automator_setup',
                                  rs_type=automator_proto.AdminSetupAutomatorResponse)
        if rs.success:
            logging.info('Automator \"%s\" is setup', automator.name)
        else:
            logging.warning('Automator \"%s\" setup step #2 error: %s', automator.name, rs.message)


class AutomatorDeleteCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_delete_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        rq = automator_proto.AdminDeleteAutomatorRequest()
        rq.automatorId = automator.automatorId
        _ = api.communicate_rest(params, rq, 'automator/automator_delete', rs_type=automator_proto.AdminResponse)
        logging.info('Automator %s deleted', automator.automatorId)
        params.automators = None


class AutomatorActionCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return automator_action_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        action = kwargs.get('action') or ''
        if action == 'reset':
            rq = automator_proto.AdminDeleteAutomatorRequest()
            endpoint = 'automator/automator_full_reset'
        elif action == 'enable' or action == 'disable':
            rq = automator_proto.AdminEnableAutomatorRequest()
            rq.enabled = action == 'enable'
            endpoint = 'automator/automator_enable'
        elif action == 'init':
            rq = automator_proto.AdminInitializeAutomatorRequest()
            endpoint = 'automator/automator_initialize'
        else:
            logging.warning('Unsupported automator action \"%s\"', action)
            return
        rq.automatorId = automator.automatorId
        rs = api.communicate_rest(params, rq, endpoint, rs_type=automator_proto.AdminResponse)
        if rs.automatorInfo:
            self.dump_automator(rs.automatorInfo[0])
            params.automators = None
        if not rs.success:
            logging.warning(rs.message)


class AutomatorLogCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_action_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        action = kwargs.get('action') or ''
        if action == 'log':
            rq = automator_proto.AdminAutomatorLogRequest()
            endpoint = 'automator/automator_log_get'
        elif action == 'log-clear':
            rq = automator_proto.AdminAutomatorLogClearRequest()
            endpoint = 'automator/automator_log_clear'
        else:
            logging.warning('Unsupported automator action \"%s\"', action)
            return
        rq.automatorId = automator.automatorId
        rs = api.communicate_rest(params, rq, endpoint, rs_type=automator_proto.AdminResponse)
        if rs.automatorInfo:
            endpoint = rs.automatorInfo[0]
            if endpoint.logEntries:
                logging.info('\n{0:>32s}\n'.format('Automator Log'))
                for log in endpoint.logEntries:
                    logging.info('<{0}> {1} - {2}'.format(log.messageLevel, log.serverTime, log.message))
            else:
                logging.info('\nNo log entries found.')


class AutomatorCertificateCommand(Command):
    def get_parser(self):
        return automator_certificate_parser

    def execute(self, params, **kwargs):
        filename = kwargs.get('file')
        if not filename:
            raise error.CommandError('automator-certificate', 'PFX file name parameter is required')

        filepath = os.path.expanduser(filename)
        if not os.path.isfile(filepath):
            raise error.CommandError('automator-certificate', f'File \"{filename}\" cannot be found')

        password = getpass.getpass(prompt='...' + 'Certificate Passphrase'.rjust(30) + ': ', stream=None)
        if not password:
            password = None

        with open(filepath, 'rb') as pfx:
            _, cert, chain = pkcs12.load_key_and_certificates(pfx.read(), password.encode('utf-8'))

        def print_cert(title, c):
            print('')
            print(f'{"":>16} {title}')
            print(f'{"Subject:":>16} {c.subject.rfc4514_string()}')
            print(f'{"Issuer:":>16} {c.issuer.rfc4514_string()}')
            print(f'{"Valid from:":>16} {c.not_valid_before}')
            print(f'{"Valid until:":>16} {c.not_valid_after}')

        print_cert('MAIN CERTIFICATE', cert)
        if chain:
            for i in range(len(chain)):
                print_cert(f'ADDITIONAL CERTIFICATE #{i+1}', chain[i])

    def is_authorised(self):
        return False

    


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper SDK
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import collections
from typing import Dict

FieldType = collections.namedtuple('FieldType', ['name', 'value', 'description'])

FieldTypes = {x[0]: x for x in (
    FieldType('text', '', 'plain text'),
    FieldType('url', '', 'url string, can be clicked'),
    FieldType('multiline', '', 'multiline text'),
    FieldType('json', '', 'json text; only validated data persisted'),
    FieldType('fileRef', '', 'reference to the file field on another record'),
    FieldType('email', '', 'valid email address plus tag'),
    FieldType('secret', '', 'the field value is masked'),
    FieldType('otp', '', 'captures the seed, displays QR code'),
    FieldType('login', '', 'Login field, detected as the website login for browser extension or KFFA.'),
    FieldType('password', '', 'Field value is masked and allows for generation. Also complexity enforcements.'),


    FieldType('host', {'hostName': '', 'port': ''}, 'multiple fields to capture host information'),
    FieldType('phone', {'region': '', 'number': '', 'ext': '', 'type': ''}, 'numbers and symbols only plus tag'),
    FieldType('name', {'first': '', 'middle': '', 'last': ''}, 'multiple fields to capture name'),
    FieldType('address', {'street1': '', 'street2': '', 'city': '', 'state': '', 'zip': '', 'country': ''},
                  'multiple fields to capture address'),
    FieldType('securityQuestion', {'question': '', 'answer': ''}, 'Security Question and Answer'),
    FieldType('paymentCard', {'cardNumber': '', 'cardExpirationDate': '', 'cardSecurityCode': ''},
              'Field consisting of validated card number, expiration date and security code.'),
    FieldType('bankAccount', {'accountType': '', 'routingNumber': '', 'accountNumber': '', 'otherType': ''},
              'bank account information'),
    FieldType('privateKey', {'publicKey': '', 'privateKey': ''}, 'private and/or public keys in ASN.1 format'),

    FieldType('addressRef', '', 'reference to the address field on another record'),
    FieldType('cardRef', '', 'reference to the card record type'),
    FieldType('date', 0, 'calendar date with validation, stored as unix milliseconds'),

    FieldType('checkbox', False, 'on/off checkbox'),
    FieldType('dropdown', '', 'list of text choices'),
    FieldType('schedule', {'type': '', 'time': '', 'month': ''}, 'schedule information'),
    FieldType('passkey', {'privateKey': {}, 'credentialId': '', 'signCount': 0, 'userId': '', 'relyingParty': '',
                          'username': '', 'createdDate': 0}, 'passwordless login passkey'),
    FieldType('script', {'fileRef': '', 'command': '', 'recordRef': [], }, 'Post rotation script'),
    FieldType('recordRef', '', 'reference to other record'),
    FieldType('appFiller', {'macroSequence': '', 'applicationTitle': '', 'contentFilter': ''}, 'Native Application Filler'),
    FieldType('pamResources', {'controllerUid': '', 'folderUid': '', 'resourceRef': []}, 'PAM resources'),
    FieldType('pamSettings', {'allowSupplyHost': False, 'connection': {}, 'portForward': {}}, 'PAM Settings'),
    FieldType('pamRemoteBrowserSettings', {'connection': {}}, 'RBI Settings'),
    FieldType('rbiUrl', '', 'RBI URL'),

)}   # type: Dict[str, FieldType]


class Multiple:
    Never = 0
    Optional = 1
    Always = 2


RecordField = collections.namedtuple('RecordField', ['name', 'type', 'multiple'])
RecordFields = {x[0]: x for x in (
    RecordField('login', 'login', Multiple.Never),
    RecordField('password', 'password', Multiple.Never),
    RecordField('company', 'text', Multiple.Never),
    RecordField('licenseNumber', 'multiline', Multiple.Never),
    RecordField('accountNumber', 'text', Multiple.Never),
    RecordField('bankAccount', 'bankAccount', Multiple.Never),
    RecordField('note', 'multiline', Multiple.Never),
    RecordField('oneTimeCode', 'otp', Multiple.Never),
    RecordField('keyPair', 'privateKey', Multiple.Never),
    RecordField('pinCode', 'secret', Multiple.Never),
    RecordField('expirationDate', 'date', Multiple.Never),
    RecordField('birthDate', 'date', Multiple.Never),
    RecordField('securityQuestion', 'securityQuestion', Multiple.Always),
    RecordField('fileRef', 'fileRef', Multiple.Always),

    RecordField('pamResources', 'pamResources', Multiple.Never),
    RecordField('pamHostname', 'host', Multiple.Never),
    RecordField('databaseType', 'dropdown', Multiple.Never),
    RecordField('directoryType', 'dropdown', Multiple.Never),
    RecordField('pamSettings', 'pamSettings', Multiple.Never),
    RecordField('pamRemoteBrowserSettings', 'pamRemoteBrowserSettings', Multiple.Never),
    RecordField('rbiUrl', 'rbiUrl', Multiple.Never),

)}   # type: Dict[str, RecordField]

for ft in FieldTypes.values():
    if ft.name not in RecordFields:
        RecordFields[ft.name] = RecordField(ft.name, ft.name, Multiple.Optional)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from flask import Flask
import logging
from werkzeug.middleware.proxy_fix import ProxyFix
from .decorators.security import limiter
from .api.routes import init_routes
from .decorators.logging import logger

    
def create_app():
    """Create and configure the Keeper Commander Service."""
    logger.debug("Initializing Keeper Commander Service")

    # Custom logging filter to replace SSL handshake errors with user-friendly message
    class SSLHandshakeFilter(logging.Filter):
        def filter(self, record):
            # Replace "Bad request version" errors with a clearer message
            if hasattr(record, 'getMessage'):
                message = record.getMessage()
                if "Bad request version" in message and any(ord(c) > 127 for c in message):
                    # Replace the ugly SSL handshake error with a user-friendly message
                    record.msg = "HTTPS request received but HTTPS protocol is not enabled on this service"
                    record.args = ()
            return True
    
    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)
    log.addFilter(SSLHandshakeFilter())
    
    app = Flask(__name__)
    app.wsgi_app = ProxyFix(app.wsgi_app, x_proto=1, x_host=1, x_prefix=1)
    
    try:
        logger.debug("Configuring rate limiter")
        limiter.init_app(app)
        
        logger.debug("Initializing API routes")
        init_routes(app)
        
        print("Keeper Commander Service initialization complete")
        return app
        
    except Exception as e:
        logger.error(f"Failed to initialize Keeper Commander Service: {str(e)}")
        raise


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import io
import json
import logging
import os.path
import pathlib
import sys
import zipfile

from typing import List, Optional, Any, Dict
from contextlib import contextmanager

from .. import imp_exp
from ..importer import (BaseFileImporter, BaseExporter, Record, RecordField, RecordSchemaField, RecordReferences,
                        Folder, SharedFolder, Permission, Team, Attachment,
                        BaseDownloadMembership, BaseDownloadRecordType, RecordType, RecordTypeField)
from ... import api, utils, record_types
from ...proto import enterprise_pb2


class KeeperJsonMixin:
    @staticmethod
    def json_to_record(j_record):   # type: (Dict[str, Any]) -> Optional[Record]
        record = Record()
        record.uid = j_record.get('uid')
        if '$type' in j_record:
            record.type = j_record['$type']
        record.title = j_record.get('title') or ''
        record.login = j_record.get('login') or ''
        record.password = j_record.get('password') or ''
        record.login_url = j_record.get('login_url') or ''
        record.notes = j_record.get('notes') or ''
        if 'last_modified' in j_record:
            lm = j_record['last_modified']
            if isinstance(lm, int):
                record.last_modified = lm
        custom_fields = j_record.get('custom_fields')
        if type(custom_fields) is dict:
            for name in custom_fields:
                value = custom_fields[name]
                if name[0] == '$':
                    pos = name.find(':')
                    if pos > 0:
                        field_type = name[1:pos].strip()
                        field_name = name[pos+1:].strip()
                    else:
                        field_type = name[1:].strip()
                        field_name = ''
                else:
                    field_type = ''
                    field_name = name
                if len(field_name) >= 2 and field_name[-2] == ':' and field_name[-1].isdigit():
                    field_name = field_name[:-2]

                is_multiple = False
                if field_type:
                    ft = record_types.RecordFields.get(field_type)
                    if ft:
                        is_multiple = ft.multiple != record_types.Multiple.Never

                if isinstance(value, list) and not is_multiple:
                    for v in value:
                        field = RecordField()
                        field.type = field_type
                        field.label = field_name
                        field.value = v
                        record.fields.append(field)
                else:
                    field = RecordField()
                    field.type = field_type
                    field.label = field_name
                    field.value = value
                    record.fields.append(field)
        if 'schema' in j_record:
            record.schema = []
            for s in j_record['schema']:
                if s.endswith(':1'):
                    s = s[:-2]
                pos = s.find(':')
                if pos > 0:
                    schema_ref = s[0:pos].strip()
                    schema_label = s[pos+1:].strip()
                else:
                    schema_ref = s
                    schema_label = ''
                if schema_ref[0] == '$':
                    schema_ref = schema_ref[1:]

                sf = RecordSchemaField()
                sf.ref = schema_ref
                sf.label = schema_label
                record.schema.append(sf)
        if 'references' in j_record:
            record.references = []
            for ref_name in j_record['references']:
                if not ref_name:
                    continue
                ref_value = j_record['references'][ref_name]
                if not ref_value:
                    continue
                if type(ref_value) != list:
                    ref_value = [ref_value]
                ref_type = ref_name
                ref_label = ''
                pos = ref_name.find(':')
                if pos > 0:
                    ref_type = ref_name[1:pos].strip()
                    ref_label = ref_name[pos+1].strip()
                if ref_type[0] == '$':
                    ref_type = ref_type[1:]
                rr = RecordReferences()
                rr.type = ref_type
                rr.label = ref_label
                rr.uids.extend(ref_value)
                record.references.append(rr)
        if 'folders' in j_record:
            record.folders = []
            for f in j_record['folders']:
                folder = Folder()
                folder.domain = f.get('shared_folder')
                folder.path = f.get('folder')
                folder.can_edit = f.get('can_edit')
                folder.can_share = f.get('can_share')
                record.folders.append(folder)
        return record


class ZipAttachment(Attachment):
    def __init__(self, zip_filename, file_uid):
        super().__init__()
        self.zip_filename = zip_filename
        self.file_uid = file_uid

    @contextmanager
    def open(self):
        with zipfile.ZipFile(self.zip_filename, mode='r') as zf:
            yield io.BytesIO(zf.read(f'files/{self.file_uid}'))

    def prepare(self):
        try:
            with zipfile.ZipFile(self.zip_filename, mode='r') as zf:
                try:
                    zi = zf.getinfo(f'files/{self.file_uid}')
                    self.size = zi.file_size
                except KeyError:
                    logging.debug('ZipAttachment: file \"%s\" not found', self.file_uid)
        except Exception as e:
            logging.debug('ZipAttachment: %s', e)
            self.size = 0


class KeeperJsonImporter(BaseFileImporter, KeeperJsonMixin):
    def do_import(self, filename, **kwargs):
        users_only = kwargs.get('users_only') or False
        if not os.path.isfile(filename):
            zip_name = pathlib.Path(filename).with_suffix('.zip').name
            if os.path.isfile(zip_name):
                if zipfile.is_zipfile(zip_name):
                    filename = zip_name
        file_path = pathlib.Path(filename)
        zip_archive = file_path.suffix == '.zip'
        if zip_archive:
            with zipfile.ZipFile(filename, 'r') as zf:
                export = json.loads(zf.read('export.json'))
        else:
            with open(filename, "r", encoding='utf-8') as jf:
                export = json.load(jf)

        records = None
        folders = None
        teams = None
        if type(export) == list:
            records = export

        elif type(export) == dict:
            records = export.get('records')
            folders = export.get('shared_folders')
            teams = export.get('teams') if users_only else None

        if folders:
            for shf in folders:
                fol = SharedFolder()
                fol.uid = shf.get('uid')
                fol.path = shf.get('path')
                fol.manage_records = shf.get('manage_records') or False
                fol.manage_users = shf.get('manage_users') or False
                fol.can_edit = shf.get('can_edit') or False
                fol.can_share = shf.get('can_share') or False
                if users_only and 'permissions' in shf:
                    fol.permissions = []
                    permissions = shf['permissions']
                    if not isinstance(permissions, list):
                        permissions = [permissions]
                    for perm in permissions:
                        if isinstance(perm, dict):
                            p = Permission()
                            p.uid = perm.get('uid')
                            p.name = perm.get('name')
                            if p.uid or p.name:
                                p.manage_records = perm.get('manage_records') or False
                                p.manage_users = perm.get('manage_users') or False
                                fol.permissions.append(p)

                yield fol

        if isinstance(teams, list):
            for t in teams:
                team = Team()
                team.name = t.get('name')
                if team.name:
                    team.uid = t.get('uid')
                    ms = t.get('members')
                    if isinstance(ms, list):
                        team.members = [x for x in ms if isinstance(x, str) and len(x) > 3]
                    yield team

        if not users_only and records:
            for r in records:
                record = KeeperJsonMixin.json_to_record(r)
                if zip_archive and 'attachments' in r:
                    attachments = r['attachments']
                    record.attachments = []
                    if isinstance(attachments, list):
                        for atta in attachments:
                            file_uid = atta.get('file_uid')
                            a = ZipAttachment(filename, file_uid)
                            a.name = atta.get('name') or file_uid
                            a.mime = atta.get('mime')
                            record.attachments.append(a)
                yield record

    def extension(self):
        return 'json'


class KeeperJsonExporter(BaseExporter):
    def do_export(self, filename, items, zip_archive=None, **kwargs):
        shared_folders = []     # type: List[SharedFolder]
        records = []            # type: List[Record]
        teams = []              # type: List[Team]

        if zip_archive is True and not filename:
            raise ValueError('Please provide zip archive file name')

        for item in items:
            if isinstance(item, Record):
                records.append(item)
            elif isinstance(item, SharedFolder):
                shared_folders.append(item)
            elif isinstance(item, Team):
                teams.append(item)

        ts = []
        for t in teams:
            team = {
                'name': t.name,
            }
            if t.uid:
                team['uid'] = t.uid
            if t.members:
                team['members'] = [x for x in t.members]
            ts.append(team)

        sfs = []
        for sf in shared_folders:
            sfo = {
                'path': sf.path,
            }
            if sf.uid:
                sfo['uid'] = sf.uid
            if sf.manage_users is not None:
                sfo['manage_users'] = sf.manage_users
            if sf.manage_records is not None:
                sfo['manage_records'] = sf.manage_records
            if sf.can_edit is not None:
                sfo['can_edit'] = sf.can_edit
            if sf.can_share is not None:
                sfo['can_share'] = sf.can_share

            if sf.permissions:
                sfo['permissions'] = []
                for perm in sf.permissions:
                    po = {
                        'name': perm.name,
                        'manage_users': perm.manage_users,
                        'manage_records': perm.manage_records
                    }
                    if perm.uid:
                        po['uid'] = perm.uid
                    sfo['permissions'].append(po)
            sfs.append(sfo)

        rs = []
        atta = {}
        for r in records:
            ro = {
                'title': r.title or ''
            }
            if r.uid:
                ro['uid'] = r.uid
            if r.login:
                ro['login'] = r.login
            if r.password:
                ro['password'] = r.password
            if r.login_url:
                ro['login_url'] = r.login_url
            if r.notes:
                ro['notes'] = r.notes
            if r.type:
                ro['$type'] = r.type
            if r.uid:
                ro['uid'] = r.uid
            if isinstance(r.last_modified, int) and r.last_modified > 0:
                ro['last_modified'] = int(r.last_modified / 1000)

            if r.fields:
                ro['custom_fields'] = {}
                for field in r.fields:
                    if not field.type and field.label and field.label.startswith('$'):
                        field.type = 'text'
                    if field.type and field.label:
                        name = f'${field.type}:{field.label}'
                    elif field.type:
                        name = f'${field.type}'
                    else:
                        name = field.label or '<No Name>'
                    value = field.value
                    if name in ro['custom_fields']:
                        orig_value = ro['custom_fields'][name]
                        if orig_value:
                            orig_value = orig_value if type(orig_value) is list else [orig_value]
                        else:
                            orig_value = []
                        if value:
                            orig_value.append(value)
                        value = orig_value
                    ro['custom_fields'][name] = value

            if r.schema:
                ro['schema'] = []
                for rsf in r.schema:
                    name = f'${rsf.ref}'
                    if rsf.label:
                        name += f':{rsf.label}'
                    ro['schema'].append(name)

            if r.references:
                ro['references'] = {}
                for ref in r.references:
                    ref_name = f'${ref.type}:{ref.label}' if ref.type and ref.label else f'${ref.type}' if ref.type else ref.label or ''
                    refs = ro['references'].get(ref_name)
                    if refs is None:
                        refs = []
                        ro['references'][ref_name] = refs
                    refs.extend(ref.uids)

            if r.folders:
                ro['folders'] = []
                for folder in r.folders:
                    if folder.domain or folder.path:
                        fo = {}
                        ro['folders'].append(fo)
                        if folder.domain:
                            fo['shared_folder'] = folder.domain
                        if folder.path:
                            fo['folder'] = folder.path
                        if folder.can_edit:
                            fo['can_edit'] = True
                        if folder.can_share:
                            fo['can_share'] = True

            if r.attachments and zip_archive:
                ro['attachments'] = []
                for at in r.attachments:
                    file_uid = at.file_uid or utils.generate_uid()
                    atta[file_uid] = at
                    a = {
                        'file_uid': file_uid,
                        'name': at.name
                    }
                    if at.mime:
                        a['mime'] = at.mime
                    ro['attachments'].append(a)

            rs.append(ro)

        jo = {}
        if ts:
            jo['teams'] = ts
        if sfs:
            jo['shared_folders'] = sfs
        if rs:
            jo['records'] = rs

        if zip_archive and filename:
            zip_name = pathlib.Path(filename).with_suffix('.zip').name
            with zipfile.ZipFile(zip_name, mode='w', compresslevel=zipfile.ZIP_DEFLATED) as zf:
                f = json.dumps(jo, indent=2, ensure_ascii=False)
                zf.writestr('export.json', f)
                total = len(atta)
                if total > 0:
                    logging.info('Downloading attachments...')
                    i = 1
                    for file_uid, at in atta.items():
                        logging.info(f'{i:>3} of {total:3} {at.name}')
                        i += 1
                        with at.open() as fs:
                            data = fs.read()
                            if data:
                                zf.writestr(f'files/{file_uid}', data)
        elif filename:
            with open(filename, mode="w", encoding='utf-8') as f:
                json.dump(jo, f, indent=2, ensure_ascii=False)
        else:
            json.dump(jo, sys.stdout, indent=2, ensure_ascii=False)
            print('')

    def has_shared_folders(self):
        return True

    def has_attachments(self):
        return True

    def extension(self):
        return 'json'

    def supports_stdout(self):
        return True

    def supports_v3_record(self):
        return True


class KeeperMembershipDownload(BaseDownloadMembership):
    def download_membership(self, params, **kwargs):
        teams = {}
        if params.shared_folder_cache:
            for shared_folder_uid in params.shared_folder_cache:
                shared_folder = api.get_shared_folder(params, shared_folder_uid)
                sf = SharedFolder()
                sf.uid = shared_folder.shared_folder_uid
                sf.path = imp_exp.get_folder_path(params, shared_folder.shared_folder_uid)
                sf.manage_users = shared_folder.default_manage_users
                sf.manage_records = shared_folder.default_manage_records
                sf.can_edit = shared_folder.default_can_edit
                sf.can_share = shared_folder.default_can_share
                sf.permissions = []
                if shared_folder.teams:
                    for team in shared_folder.teams:
                        perm = Permission()
                        perm.uid = team['team_uid']
                        perm.name = team['name']
                        perm.manage_users = team.get('manage_users', False)
                        perm.manage_records = team.get('manage_records', False)
                        teams[perm.uid] = perm.name
                        sf.permissions.append(perm)
                if shared_folder.users:
                    for user in shared_folder.users:
                        perm = Permission()
                        perm.name = user['username']
                        perm.manage_users = user.get('manage_users', False)
                        perm.manage_records = user.get('manage_records', False)
                        sf.permissions.append(perm)
                yield sf

        folders_only = kwargs.get('folders_only') is True
        if folders_only is True:
            return

        enterprise_teams = {}    # type: Dict[int, List[str]]
        if params.enterprise:
            users = {x['enterprise_user_id']: x['username'] for x in params.enterprise.get('users', [])
                     if x.get('status') == 'active'}
            if 'team_users' in params.enterprise:
                for tu in params.enterprise['team_users']:
                    team_uid = tu.get('team_uid')
                    user_id = tu.get('enterprise_user_id')
                    if team_uid and user_id:
                        if user_id in users:
                            if team_uid not in enterprise_teams:
                                enterprise_teams[team_uid] = []
                            enterprise_teams[team_uid].append(users[user_id])

        if teams and params.enterprise_ec_key:
            for team_uid in teams:
                t = Team()
                t.uid = team_uid
                t.name = teams[team_uid]
                if team_uid in enterprise_teams:
                    t.members = list(enterprise_teams[team_uid])
                else:
                    rq = enterprise_pb2.GetTeamMemberRequest()
                    rq.teamUid = utils.base64_url_decode(team_uid)
                    rs = api.communicate_rest(params, rq, 'vault/get_team_members', rs_type=enterprise_pb2.GetTeamMemberResponse)
                    t.members = [x.email for x in rs.enterpriseUser]
                yield t


class KeeperRecordTypeDownload(BaseDownloadRecordType):
    def download_record_type(self, params, **kwargs):
        if params.record_type_cache:
            for rt_id, rts in params.record_type_cache.items():
                if rt_id <= 1000:
                    continue
                try:
                    rto = json.loads(rts)
                    if '$id' in rto and 'fields' in rto:
                        rt = RecordType()
                        rt.name = rto['$id']
                        rt.description = rto.get('description') or ''
                        for field in rto['fields']:
                            rtf = RecordTypeField()
                            rtf.type = field['$ref']
                            if 'label' in field:
                                rtf.label = field['label']
                            if 'required' in field:
                                req = field['required']
                                if req:
                                    rtf.required = True
                            rt.fields.append(rtf)
                        yield rt
                except:
                    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: BI.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x08\x42I.proto\x12\x02\x42I\x1a\x1cgoogle/protobuf/struct.proto\"f\n\x1bValidateSessionTokenRequest\x12\x1d\n\x15\x65ncryptedSessionToken\x18\x01 \x01(\x0c\x12\x1c\n\x14returnMcEnterpiseIds\x18\x02 \x01(\x08\x12\n\n\x02ip\x18\x03 \x01(\t\"\xda\x02\n\x1cValidateSessionTokenResponse\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x0e\n\x06userId\x18\x02 \x01(\x05\x12\x18\n\x10\x65nterpriseUserId\x18\x03 \x01(\x03\x12\x37\n\x06status\x18\x04 \x01(\x0e\x32\'.BI.ValidateSessionTokenResponse.Status\x12\x15\n\rstatusMessage\x18\x05 \x01(\t\x12\x17\n\x0fmcEnterpriseIds\x18\x06 \x03(\x05\x12\x18\n\x10hasMSPPermission\x18\x07 \x01(\x08\x12\x1e\n\x16\x64\x65letedMcEnterpriseIds\x18\x08 \x03(\x05\"[\n\x06Status\x12\t\n\x05VALID\x10\x00\x12\r\n\tNOT_VALID\x10\x01\x12\x0b\n\x07\x45XPIRED\x10\x02\x12\x0e\n\nIP_BLOCKED\x10\x03\x12\x1a\n\x16INVALID_CLIENT_VERSION\x10\x04\"\x1b\n\x19SubscriptionStatusRequest\"\xa6\x03\n\x1aSubscriptionStatusResponse\x12$\n\x0b\x61utoRenewal\x18\x01 \x01(\x0b\x32\x0f.BI.AutoRenewal\x12/\n\x14\x63urrentPaymentMethod\x18\x02 \x01(\x0b\x32\x11.BI.PaymentMethod\x12\x14\n\x0c\x63heckoutLink\x18\x03 \x01(\t\x12\x19\n\x11licenseCreateDate\x18\x04 \x01(\x03\x12\x15\n\risDistributor\x18\x05 \x01(\x08\x12\x13\n\x0bisLegacyMsp\x18\x06 \x01(\x08\x12&\n\x0clicenseStats\x18\x08 \x03(\x0b\x32\x10.BI.LicenseStats\x12\x35\n\x0egradientStatus\x18\t \x01(\x0e\x32\x1d.BI.GradientIntegrationStatus\x12\x17\n\x0fhideTrialBanner\x18\n \x01(\x08\x12\x1c\n\x14gradientLastSyncDate\x18\x0b \x01(\t\x12\x1c\n\x14gradientNextSyncDate\x18\x0c \x01(\t\x12 \n\x18isGradientMappingPending\x18\r \x01(\x08\"\xd7\x01\n\x0cLicenseStats\x12#\n\x04type\x18\x01 \x01(\x0e\x32\x15.BI.LicenseStats.Type\x12\x11\n\tavailable\x18\x02 \x01(\x05\x12\x0c\n\x04used\x18\x03 \x01(\x05\"\x80\x01\n\x04Type\x12\x18\n\x14LICENSE_STAT_UNKNOWN\x10\x00\x12\x0c\n\x08MSP_BASE\x10\x01\x12\x0f\n\x0bMC_BUSINESS\x10\x02\x12\x14\n\x10MC_BUSINESS_PLUS\x10\x03\x12\x11\n\rMC_ENTERPRISE\x10\x04\x12\x16\n\x12MC_ENTERPRISE_PLUS\x10\x05\"@\n\x0b\x41utoRenewal\x12\x0e\n\x06nextOn\x18\x01 \x01(\x03\x12\x10\n\x08\x64\x61ysLeft\x18\x02 \x01(\x05\x12\x0f\n\x07isTrial\x18\x03 \x01(\x08\"\x84\x04\n\rPaymentMethod\x12$\n\x04type\x18\x01 \x01(\x0e\x32\x16.BI.PaymentMethod.Type\x12$\n\x04\x63\x61rd\x18\x02 \x01(\x0b\x32\x16.BI.PaymentMethod.Card\x12$\n\x04sepa\x18\x03 \x01(\x0b\x32\x16.BI.PaymentMethod.Sepa\x12(\n\x06paypal\x18\x04 \x01(\x0b\x32\x18.BI.PaymentMethod.Paypal\x12\x15\n\rfailedBilling\x18\x05 \x01(\x08\x12(\n\x06vendor\x18\x06 \x01(\x0b\x32\x18.BI.PaymentMethod.Vendor\x12\x36\n\rpurchaseOrder\x18\x07 \x01(\x0b\x32\x1f.BI.PaymentMethod.PurchaseOrder\x1a$\n\x04\x43\x61rd\x12\r\n\x05last4\x18\x01 \x01(\t\x12\r\n\x05\x62rand\x18\x02 \x01(\t\x1a&\n\x04Sepa\x12\r\n\x05last4\x18\x01 \x01(\t\x12\x0f\n\x07\x63ountry\x18\x02 \x01(\t\x1a\x08\n\x06Paypal\x1a\x16\n\x06Vendor\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1d\n\rPurchaseOrder\x12\x0c\n\x04name\x18\x01 \x01(\t\"O\n\x04Type\x12\x08\n\x04\x43\x41RD\x10\x00\x12\x08\n\x04SEPA\x10\x01\x12\n\n\x06PAYPAL\x10\x02\x12\x08\n\x04NONE\x10\x03\x12\n\n\x06VENDOR\x10\x04\x12\x11\n\rPURCHASEORDER\x10\x05\"\x1f\n\x1dSubscriptionMspPricingRequest\"\\\n\x1eSubscriptionMspPricingResponse\x12\x19\n\x06\x61\x64\x64ons\x18\x02 \x03(\x0b\x32\t.BI.Addon\x12\x1f\n\tfilePlans\x18\x03 \x03(\x0b\x32\x0c.BI.FilePlan\"\x1e\n\x1cSubscriptionMcPricingRequest\"|\n\x1dSubscriptionMcPricingResponse\x12\x1f\n\tbasePlans\x18\x01 \x03(\x0b\x32\x0c.BI.BasePlan\x12\x19\n\x06\x61\x64\x64ons\x18\x02 \x03(\x0b\x32\t.BI.Addon\x12\x1f\n\tfilePlans\x18\x03 \x03(\x0b\x32\x0c.BI.FilePlan\".\n\x08\x42\x61sePlan\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x16\n\x04\x63ost\x18\x02 \x01(\x0b\x32\x08.BI.Cost\"C\n\x05\x41\x64\x64on\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x16\n\x04\x63ost\x18\x02 \x01(\x0b\x32\x08.BI.Cost\x12\x16\n\x0e\x61mountConsumed\x18\x03 \x01(\x03\".\n\x08\x46ilePlan\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x16\n\x04\x63ost\x18\x02 \x01(\x0b\x32\x08.BI.Cost\"\xbf\x01\n\x04\x43ost\x12\x0e\n\x06\x61mount\x18\x03 \x01(\x01\x12%\n\tamountPer\x18\x04 \x01(\x0e\x32\x12.BI.Cost.AmountPer\x12\x1e\n\x08\x63urrency\x18\x05 \x01(\x0e\x32\x0c.BI.Currency\"`\n\tAmountPer\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05MONTH\x10\x01\x12\x0e\n\nUSER_MONTH\x10\x02\x12\x17\n\x13USER_CONSUMED_MONTH\x10\x03\x12\x12\n\x0e\x45NDPOINT_MONTH\x10\x04\"\\\n\x14InvoiceSearchRequest\x12\x0c\n\x04size\x18\x01 \x01(\x05\x12\x17\n\x0fstartingAfterId\x18\x02 \x01(\x05\x12\x1d\n\x15\x61llInvoicesUnfiltered\x18\x03 \x01(\x08\"6\n\x15InvoiceSearchResponse\x12\x1d\n\x08invoices\x18\x01 \x03(\x0b\x32\x0b.BI.Invoice\"\xbe\x02\n\x07Invoice\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x15\n\rinvoiceNumber\x18\x02 \x01(\t\x12\x13\n\x0binvoiceDate\x18\x03 \x01(\x03\x12\x14\n\x0clicenseCount\x18\x04 \x01(\x05\x12#\n\ttotalCost\x18\x05 \x01(\x0b\x32\x10.BI.Invoice.Cost\x12%\n\x0binvoiceType\x18\x06 \x01(\x0e\x32\x10.BI.Invoice.Type\x1a\x36\n\x04\x43ost\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x01\x12\x1e\n\x08\x63urrency\x18\x02 \x01(\x0e\x32\x0c.BI.Currency\"a\n\x04Type\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x07\n\x03NEW\x10\x01\x12\x0b\n\x07RENEWAL\x10\x02\x12\x0b\n\x07UPGRADE\x10\x03\x12\x0b\n\x07RESTORE\x10\x04\x12\x0f\n\x0b\x41SSOCIATION\x10\x05\x12\x0b\n\x07OVERAGE\x10\x06\"\x1a\n\x18VaultInvoicesListRequest\"?\n\x19VaultInvoicesListResponse\x12\"\n\x08invoices\x18\x01 \x03(\x0b\x32\x10.BI.VaultInvoice\"\x8f\x01\n\x0cVaultInvoice\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x15\n\rinvoiceNumber\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x61teCreated\x18\x03 \x01(\x03\x12\x1f\n\x05total\x18\x04 \x01(\x0b\x32\x10.BI.Invoice.Cost\x12&\n\x0cpurchaseType\x18\x05 \x01(\x0e\x32\x10.BI.Invoice.Type\"/\n\x16InvoiceDownloadRequest\x12\x15\n\rinvoiceNumber\x18\x01 \x01(\t\"9\n\x17InvoiceDownloadResponse\x12\x0c\n\x04link\x18\x01 \x01(\t\x12\x10\n\x08\x66ileName\x18\x02 \x01(\t\"8\n\x1fVaultInvoiceDownloadLinkRequest\x12\x15\n\rinvoiceNumber\x18\x01 \x01(\t\"B\n VaultInvoiceDownloadLinkResponse\x12\x0c\n\x04link\x18\x01 \x01(\t\x12\x10\n\x08\x66ileName\x18\x02 \x01(\t\"<\n\x1dReportingDailySnapshotRequest\x12\r\n\x05month\x18\x01 \x01(\x05\x12\x0c\n\x04year\x18\x02 \x01(\x05\"v\n\x1eReportingDailySnapshotResponse\x12#\n\x07records\x18\x01 \x03(\x0b\x32\x12.BI.SnapshotRecord\x12/\n\rmcEnterprises\x18\x02 \x03(\x0b\x32\x18.BI.SnapshotMcEnterprise\"\xd7\x01\n\x0eSnapshotRecord\x12\x0c\n\x04\x64\x61te\x18\x01 \x01(\x03\x12\x16\n\x0emcEnterpriseId\x18\x02 \x01(\x05\x12\x17\n\x0fmaxLicenseCount\x18\x04 \x01(\x05\x12\x19\n\x11maxFilePlanTypeId\x18\x05 \x01(\x05\x12\x15\n\rmaxBasePlanId\x18\x06 \x01(\x05\x12(\n\x06\x61\x64\x64ons\x18\x07 \x03(\x0b\x32\x18.BI.SnapshotRecord.Addon\x1a*\n\x05\x41\x64\x64on\x12\x12\n\nmaxAddonId\x18\x01 \x01(\x05\x12\r\n\x05units\x18\x02 \x01(\x03\"0\n\x14SnapshotMcEnterprise\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\"\x16\n\x14MappingAddonsRequest\"\\\n\x15MappingAddonsResponse\x12\x1f\n\x06\x61\x64\x64ons\x18\x01 \x03(\x0b\x32\x0f.BI.MappingItem\x12\"\n\tfilePlans\x18\x02 \x03(\x0b\x32\x0f.BI.MappingItem\"\'\n\x0bMappingItem\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\"1\n\x1aGradientValidateKeyRequest\x12\x13\n\x0bgradientKey\x18\x01 \x01(\t\"?\n\x1bGradientValidateKeyResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"D\n\x13GradientSaveRequest\x12\x13\n\x0bgradientKey\x18\x01 \x01(\t\x12\x18\n\x10\x65nterpriseUserId\x18\x02 \x01(\x03\"g\n\x14GradientSaveResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12-\n\x06status\x18\x02 \x01(\x0e\x32\x1d.BI.GradientIntegrationStatus\x12\x0f\n\x07message\x18\x03 \x01(\t\"1\n\x15GradientRemoveRequest\x12\x18\n\x10\x65nterpriseUserId\x18\x01 \x01(\x03\":\n\x16GradientRemoveResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"/\n\x13GradientSyncRequest\x12\x18\n\x10\x65nterpriseUserId\x18\x01 \x01(\x03\"g\n\x14GradientSyncResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12-\n\x06status\x18\x02 \x01(\x0e\x32\x1d.BI.GradientIntegrationStatus\x12\x0f\n\x07message\x18\x03 \x01(\t\"N\n\'NetPromoterScoreSurveySubmissionRequest\x12\x14\n\x0csurvey_score\x18\x01 \x01(\x05\x12\r\n\x05notes\x18\x02 \x01(\t\"*\n(NetPromoterScoreSurveySubmissionResponse\"&\n$NetPromoterScorePopupScheduleRequest\";\n%NetPromoterScorePopupScheduleResponse\x12\x12\n\nshow_popup\x18\x01 \x01(\x08\"\'\n%NetPromoterScorePopupDismissalRequest\"(\n&NetPromoterScorePopupDismissalResponse\"-\n\x11KCMLicenseRequest\x12\x18\n\x10\x65nterpriseUserId\x18\x01 \x01(\x03\"%\n\x12KCMLicenseResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"\x84\x01\n\x0c\x45ventRequest\x12 \n\teventType\x18\x01 \x01(\x0e\x32\r.BI.EventType\x12\x12\n\neventValue\x18\x02 \x01(\t\x12\x11\n\teventTime\x18\x03 \x01(\x03\x12+\n\nattributes\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\"0\n\rEventsRequest\x12\x1f\n\x05\x65vent\x18\x01 \x03(\x0b\x32\x10.BI.EventRequest\"\x87\x01\n\x16\x43ustomerCaptureRequest\x12\x0f\n\x07pageUrl\x18\x01 \x01(\t\x12\x0c\n\x04tree\x18\x02 \x01(\t\x12\x0c\n\x04hash\x18\x03 \x01(\t\x12\r\n\x05image\x18\x04 \x01(\t\x12\x14\n\x0cpageLoadTime\x18\x05 \x01(\t\x12\r\n\x05keyId\x18\x06 \x01(\t\x12\x0c\n\x04test\x18\x07 \x01(\x08\"\x19\n\x17\x43ustomerCaptureResponse*M\n\x08\x43urrency\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x07\n\x03USD\x10\x01\x12\x07\n\x03GBP\x10\x02\x12\x07\n\x03JPY\x10\x03\x12\x07\n\x03\x45UR\x10\x04\x12\x07\n\x03\x41UD\x10\x05\x12\x07\n\x03\x43\x41\x44\x10\x06*S\n\x19GradientIntegrationStatus\x12\x10\n\x0cNOTCONNECTED\x10\x00\x12\x0b\n\x07PENDING\x10\x01\x12\r\n\tCONNECTED\x10\x02\x12\x08\n\x04NONE\x10\x03*\xb9\x01\n\tEventType\x12\x1f\n\x1bUNKNOWN_TRACKING_EVENT_TYPE\x10\x00\x12\x1c\n\x18TRACKING_POPUP_DISPLAYED\x10\x01\x12\x1b\n\x17TRACKING_POPUP_ACCEPTED\x10\x02\x12\x1c\n\x18TRACKING_POPUP_DISMISSED\x10\x03\x12\x17\n\x13TRACKING_POPUP_PAID\x10\x04\x12\x19\n\x15TRACKING_PUSH_CLICKED\x10\x05\x42\x1e\n\x18\x63om.keepersecurity.protoB\x02\x42Ib\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'BI_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\002BI'
  _globals['_CURRENCY']._serialized_start=5253
  _globals['_CURRENCY']._serialized_end=5330
  _globals['_GRADIENTINTEGRATIONSTATUS']._serialized_start=5332
  _globals['_GRADIENTINTEGRATIONSTATUS']._serialized_end=5415
  _globals['_EVENTTYPE']._serialized_start=5418
  _globals['_EVENTTYPE']._serialized_end=5603
  _globals['_VALIDATESESSIONTOKENREQUEST']._serialized_start=46
  _globals['_VALIDATESESSIONTOKENREQUEST']._serialized_end=148
  _globals['_VALIDATESESSIONTOKENRESPONSE']._serialized_start=151
  _globals['_VALIDATESESSIONTOKENRESPONSE']._serialized_end=497
  _globals['_VALIDATESESSIONTOKENRESPONSE_STATUS']._serialized_start=406
  _globals['_VALIDATESESSIONTOKENRESPONSE_STATUS']._serialized_end=497
  _globals['_SUBSCRIPTIONSTATUSREQUEST']._serialized_start=499
  _globals['_SUBSCRIPTIONSTATUSREQUEST']._serialized_end=526
  _globals['_SUBSCRIPTIONSTATUSRESPONSE']._serialized_start=529
  _globals['_SUBSCRIPTIONSTATUSRESPONSE']._serialized_end=951
  _globals['_LICENSESTATS']._serialized_start=954
  _globals['_LICENSESTATS']._serialized_end=1169
  _globals['_LICENSESTATS_TYPE']._serialized_start=1041
  _globals['_LICENSESTATS_TYPE']._serialized_end=1169
  _globals['_AUTORENEWAL']._serialized_start=1171
  _globals['_AUTORENEWAL']._serialized_end=1235
  _globals['_PAYMENTMETHOD']._serialized_start=1238
  _globals['_PAYMENTMETHOD']._serialized_end=1754
  _globals['_PAYMENTMETHOD_CARD']._serialized_start=1532
  _globals['_PAYMENTMETHOD_CARD']._serialized_end=1568
  _globals['_PAYMENTMETHOD_SEPA']._serialized_start=1570
  _globals['_PAYMENTMETHOD_SEPA']._serialized_end=1608
  _globals['_PAYMENTMETHOD_PAYPAL']._serialized_start=1610
  _globals['_PAYMENTMETHOD_PAYPAL']._serialized_end=1618
  _globals['_PAYMENTMETHOD_VENDOR']._serialized_start=1620
  _globals['_PAYMENTMETHOD_VENDOR']._serialized_end=1642
  _globals['_PAYMENTMETHOD_PURCHASEORDER']._serialized_start=1644
  _globals['_PAYMENTMETHOD_PURCHASEORDER']._serialized_end=1673
  _globals['_PAYMENTMETHOD_TYPE']._serialized_start=1675
  _globals['_PAYMENTMETHOD_TYPE']._serialized_end=1754
  _globals['_SUBSCRIPTIONMSPPRICINGREQUEST']._serialized_start=1756
  _globals['_SUBSCRIPTIONMSPPRICINGREQUEST']._serialized_end=1787
  _globals['_SUBSCRIPTIONMSPPRICINGRESPONSE']._serialized_start=1789
  _globals['_SUBSCRIPTIONMSPPRICINGRESPONSE']._serialized_end=1881
  _globals['_SUBSCRIPTIONMCPRICINGREQUEST']._serialized_start=1883
  _globals['_SUBSCRIPTIONMCPRICINGREQUEST']._serialized_end=1913
  _globals['_SUBSCRIPTIONMCPRICINGRESPONSE']._serialized_start=1915
  _globals['_SUBSCRIPTIONMCPRICINGRESPONSE']._serialized_end=2039
  _globals['_BASEPLAN']._serialized_start=2041
  _globals['_BASEPLAN']._serialized_end=2087
  _globals['_ADDON']._serialized_start=2089
  _globals['_ADDON']._serialized_end=2156
  _globals['_FILEPLAN']._serialized_start=2158
  _globals['_FILEPLAN']._serialized_end=2204
  _globals['_COST']._serialized_start=2207
  _globals['_COST']._serialized_end=2398
  _globals['_COST_AMOUNTPER']._serialized_start=2302
  _globals['_COST_AMOUNTPER']._serialized_end=2398
  _globals['_INVOICESEARCHREQUEST']._serialized_start=2400
  _globals['_INVOICESEARCHREQUEST']._serialized_end=2492
  _globals['_INVOICESEARCHRESPONSE']._serialized_start=2494
  _globals['_INVOICESEARCHRESPONSE']._serialized_end=2548
  _globals['_INVOICE']._serialized_start=2551
  _globals['_INVOICE']._serialized_end=2869
  _globals['_INVOICE_COST']._serialized_start=2716
  _globals['_INVOICE_COST']._serialized_end=2770
  _globals['_INVOICE_TYPE']._serialized_start=2772
  _globals['_INVOICE_TYPE']._serialized_end=2869
  _globals['_VAULTINVOICESLISTREQUEST']._serialized_start=2871
  _globals['_VAULTINVOICESLISTREQUEST']._serialized_end=2897
  _globals['_VAULTINVOICESLISTRESPONSE']._serialized_start=2899
  _globals['_VAULTINVOICESLISTRESPONSE']._serialized_end=2962
  _globals['_VAULTINVOICE']._serialized_start=2965
  _globals['_VAULTINVOICE']._serialized_end=3108
  _globals['_INVOICEDOWNLOADREQUEST']._serialized_start=3110
  _globals['_INVOICEDOWNLOADREQUEST']._serialized_end=3157
  _globals['_INVOICEDOWNLOADRESPONSE']._serialized_start=3159
  _globals['_INVOICEDOWNLOADRESPONSE']._serialized_end=3216
  _globals['_VAULTINVOICEDOWNLOADLINKREQUEST']._serialized_start=3218
  _globals['_VAULTINVOICEDOWNLOADLINKREQUEST']._serialized_end=3274
  _globals['_VAULTINVOICEDOWNLOADLINKRESPONSE']._serialized_start=3276
  _globals['_VAULTINVOICEDOWNLOADLINKRESPONSE']._serialized_end=3342
  _globals['_REPORTINGDAILYSNAPSHOTREQUEST']._serialized_start=3344
  _globals['_REPORTINGDAILYSNAPSHOTREQUEST']._serialized_end=3404
  _globals['_REPORTINGDAILYSNAPSHOTRESPONSE']._serialized_start=3406
  _globals['_REPORTINGDAILYSNAPSHOTRESPONSE']._serialized_end=3524
  _globals['_SNAPSHOTRECORD']._serialized_start=3527
  _globals['_SNAPSHOTRECORD']._serialized_end=3742
  _globals['_SNAPSHOTRECORD_ADDON']._serialized_start=3700
  _globals['_SNAPSHOTRECORD_ADDON']._serialized_end=3742
  _globals['_SNAPSHOTMCENTERPRISE']._serialized_start=3744
  _globals['_SNAPSHOTMCENTERPRISE']._serialized_end=3792
  _globals['_MAPPINGADDONSREQUEST']._serialized_start=3794
  _globals['_MAPPINGADDONSREQUEST']._serialized_end=3816
  _globals['_MAPPINGADDONSRESPONSE']._serialized_start=3818
  _globals['_MAPPINGADDONSRESPONSE']._serialized_end=3910
  _globals['_MAPPINGITEM']._serialized_start=3912
  _globals['_MAPPINGITEM']._serialized_end=3951
  _globals['_GRADIENTVALIDATEKEYREQUEST']._serialized_start=3953
  _globals['_GRADIENTVALIDATEKEYREQUEST']._serialized_end=4002
  _globals['_GRADIENTVALIDATEKEYRESPONSE']._serialized_start=4004
  _globals['_GRADIENTVALIDATEKEYRESPONSE']._serialized_end=4067
  _globals['_GRADIENTSAVEREQUEST']._serialized_start=4069
  _globals['_GRADIENTSAVEREQUEST']._serialized_end=4137
  _globals['_GRADIENTSAVERESPONSE']._serialized_start=4139
  _globals['_GRADIENTSAVERESPONSE']._serialized_end=4242
  _globals['_GRADIENTREMOVEREQUEST']._serialized_start=4244
  _globals['_GRADIENTREMOVEREQUEST']._serialized_end=4293
  _globals['_GRADIENTREMOVERESPONSE']._serialized_start=4295
  _globals['_GRADIENTREMOVERESPONSE']._serialized_end=4353
  _globals['_GRADIENTSYNCREQUEST']._serialized_start=4355
  _globals['_GRADIENTSYNCREQUEST']._serialized_end=4402
  _globals['_GRADIENTSYNCRESPONSE']._serialized_start=4404
  _globals['_GRADIENTSYNCRESPONSE']._serialized_end=4507
  _globals['_NETPROMOTERSCORESURVEYSUBMISSIONREQUEST']._serialized_start=4509
  _globals['_NETPROMOTERSCORESURVEYSUBMISSIONREQUEST']._serialized_end=4587
  _globals['_NETPROMOTERSCORESURVEYSUBMISSIONRESPONSE']._serialized_start=4589
  _globals['_NETPROMOTERSCORESURVEYSUBMISSIONRESPONSE']._serialized_end=4631
  _globals['_NETPROMOTERSCOREPOPUPSCHEDULEREQUEST']._serialized_start=4633
  _globals['_NETPROMOTERSCOREPOPUPSCHEDULEREQUEST']._serialized_end=4671
  _globals['_NETPROMOTERSCOREPOPUPSCHEDULERESPONSE']._serialized_start=4673
  _globals['_NETPROMOTERSCOREPOPUPSCHEDULERESPONSE']._serialized_end=4732
  _globals['_NETPROMOTERSCOREPOPUPDISMISSALREQUEST']._serialized_start=4734
  _globals['_NETPROMOTERSCOREPOPUPDISMISSALREQUEST']._serialized_end=4773
  _globals['_NETPROMOTERSCOREPOPUPDISMISSALRESPONSE']._serialized_start=4775
  _globals['_NETPROMOTERSCOREPOPUPDISMISSALRESPONSE']._serialized_end=4815
  _globals['_KCMLICENSEREQUEST']._serialized_start=4817
  _globals['_KCMLICENSEREQUEST']._serialized_end=4862
  _globals['_KCMLICENSERESPONSE']._serialized_start=4864
  _globals['_KCMLICENSERESPONSE']._serialized_end=4901
  _globals['_EVENTREQUEST']._serialized_start=4904
  _globals['_EVENTREQUEST']._serialized_end=5036
  _globals['_EVENTSREQUEST']._serialized_start=5038
  _globals['_EVENTSREQUEST']._serialized_end=5086
  _globals['_CUSTOMERCAPTUREREQUEST']._serialized_start=5089
  _globals['_CUSTOMERCAPTUREREQUEST']._serialized_end=5224
  _globals['_CUSTOMERCAPTURERESPONSE']._serialized_start=5226
  _globals['_CUSTOMERCAPTURERESPONSE']._serialized_end=5251
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
from .constants import VERTICES_SORT_MAP
from .types import DiscoveryObject
import logging
import functools
import re
from typing import List, Optional, Union, TYPE_CHECKING

Logger = Union[logging.RootLogger, logging.Logger]
if TYPE_CHECKING:
    from ..keeper_dag.vertex import DAGVertex


def sort_infra_name(vertices: List[DAGVertex]) -> List[DAGVertex]:
    """
    Sort the vertices by name in ascending order.
    """

    def _sort(t1: DAGVertex, t2: DAGVertex):
        t1_name = t1.content_as_dict.get("name")
        t2_name = t2.content_as_dict.get("name")
        if t1_name < t2_name:
            return -1
        elif t1_name > t2_name:
            return 1
        else:
            return 0

    return sorted(vertices, key=functools.cmp_to_key(_sort))


def sort_infra_host(vertices: List[DAGVertex]) -> List[DAGVertex]:
    """
    Sort the vertices by host name.

    Host name should appear first in ascending order.
    IP should appear second in ascending order.

    """

    def _is_ip(host: str) -> bool:
        if re.match(r'^\d+\.\d+\.\d+\.\d+', host) is not None:
            return True
        return False

    def _make_ip_number(ip: str) -> int:
        ip_port = ip.split(":")
        parts = ip_port[0].split(".")
        value = ""
        for part in parts:
            value += part.zfill(3)
        return int(value)

    def _sort(t1: DAGVertex, t2: DAGVertex):
        t1_name = t1.content_as_dict.get("name")
        t2_name = t2.content_as_dict.get("name")

        # Both names are ip addresses
        if _is_ip(t1_name) and _is_ip(t2_name):
            t1_num = _make_ip_number(t1_name)
            t2_num = _make_ip_number(t2_name)

            if t1_num < t2_num:
                return -1
            elif t1_num > t2_num:
                return 1
            else:
                return 0

        # T1 is an IP, T2 is a host name
        elif _is_ip(t1_name) and not _is_ip(t2_name):
            return 1
        # T2 is not an IP and T2 is an IP
        elif not _is_ip(t1_name) and _is_ip(t2_name):
            return -1
        # T1 and T2 are host name
        else:
            if t1_name < t2_name:
                return -1
            elif t1_name > t2_name:
                return 1
            else:
                return 0

    return sorted(vertices, key=functools.cmp_to_key(_sort))


def sort_infra_vertices(current_vertex: DAGVertex, logger: Optional[Logger] = None) -> dict:

    if logger is None:
        logger = logging.getLogger()

    # Make a map, record type to list of vertices (of that record type)
    record_type_to_vertices_map = {k: [] for k, v in VERTICES_SORT_MAP.items()}

    # Collate the vertices into a record type lookup.
    vertices = current_vertex.has_vertices()
    logger.debug(f"  found {len(vertices)} vertices")
    for vertex in vertices:
        if vertex.active is True:
            content = DiscoveryObject.get_discovery_object(vertex)
            logger.debug(f"  * {content.description}")
    for vertex in vertices:
        if vertex.active is False:
            logger.debug("  vertex is not active")
            continue
        # We can't load into a pydantic object since Pydantic has a problem with Union type.
        # We only want the record type, so it is too much work to try to get into an object.
        content_dict = vertex.content_as_dict
        record_type = content_dict.get("record_type")
        if record_type in record_type_to_vertices_map:
            record_type_to_vertices_map[record_type].append(vertex)

    # Sort the vertices for each record type.
    for k, v in VERTICES_SORT_MAP.items():
        if v["sort"] == "sort_infra_name":
            record_type_to_vertices_map[k] = sort_infra_name(record_type_to_vertices_map[k])
        elif v["sort"] == "sort_infra_host":
            record_type_to_vertices_map[k] = sort_infra_host(record_type_to_vertices_map[k])

    return record_type_to_vertices_map



import csv
import logging
from typing import List

from keepercommander.importer.importer import Record, RecordField, RecordReferences
from keepercommander import vault
from keepercommander.importer.json import Exporter

records = []           # type: List[Record]

with open('contacts.csv', "r", encoding='utf-8-sig') as csvfile:
    reader = csv.DictReader(csvfile)
    for row in reader:   # type: dict
        r = Record()
        r.type = 'contact'
        name = [str(row.pop(x, '')).strip() for x in ('First Name', 'Middle Name', 'Last Name')]
        if any(name):
            rf = RecordField('name', '', {
                'first': name[0], 'middle': name[1], 'last': name[2]
            })
            r.fields.append(rf)
            r.title = vault.TypedField.export_name_field(rf.value)
        r.notes = str(row.pop('Notes', '')).strip()
        company = str(row.pop('Company', '')).strip()
        if company:
            rf = RecordField('text', 'company', company)
            r.fields.append(rf)
            if not r.title:
                r.title = company

        name = [str(row.pop(x, '')).strip() for x in ('Given Yomi', 'Surname Yomi')]
        if any(name):
            rf = RecordField('name', 'Yomi', {
                'first': name[0], 'middle': '', 'last': name[1]
            })
            r.fields.append(rf)

        address_refs = []
        for prefix in ('Business', 'Home', 'Other'):
            address = [str(row.pop(x, '')).strip() for x in (f'{prefix} Street',
                                                             f'{prefix} City',
                                                             f'{prefix} State',
                                                             f'{prefix} Postal Code',
                                                             f'{prefix} Country/Region')]
            if any(address):
                if address[-1] == 'USA':
                    address[-1] = 'US'
                rf = RecordField('address', '', {
                    'street1': address[0],
                    'street2': '',
                    'city': address[1],
                    'state': address[2],
                    'zip': address[3],
                    'country': address[4],
                })
                ar = Record()
                ar.title = f'{prefix}: {r.title}'
                ar.type = 'address'
                ar.fields.append(rf)
                records.append(ar)
                ar.uid = len(records)
                address_refs.append(ar.uid)
        if address_refs:
            if r.references is None:
                r.references = []
            rr = RecordReferences('address')
            rr.uids = address_refs
            r.references.append(rr)

        phones = []
        for label, value in row.items():
            if not isinstance(value, str):
                continue
            value = value.strip()
            if not value:
                continue
            if 'E-mail' in label:
                label = label.replace('Address', '').replace('  ', ' ').strip()
                if label == 'E-mail':
                    label = ''
                rf = RecordField('email', '', value)
                r.fields.append(rf)
            elif 'Phone' in label:
                label = label.replace('Phone', '').replace('  ', ' ').strip()
                rv = {
                    'type': label,
                    'region': '',
                    'number': value,
                    'ext': '',
                }
                phones.append(rv)
            elif 'Web Page' in label:
                rf = RecordField('url', label, value)
                r.fields.append(rf)
            else:
                rf = RecordField('text', label, value)
                r.fields.append(rf)
        if phones:
            rf = RecordField('phone', '', phones)
            r.fields.append(rf)
        records.append(r)
        r.uid = len(records)
        if not r.title:
            r.title = f'Contact #{r.uid}'

if records:
    exporter = Exporter()
    exporter.execute('contacts.json', records)
    logging.info('Exported %d records', len(records))





